#ifndef BENCHMARKS_CPP_RAW_RAW_BENCH_H_
#define BENCHMARKS_CPP_RAW_RAW_BENCH_H_

#include <memory>

#include "benchmarks/cpp/bench.h"

std::unique_ptr<Bench> NewRawBench();

#endif  // BENCHMARKS_CPP_RAW_RAW_BENCH_H_