#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from conans import CMake, ConanFile


class TestPackageConan(ConanFile):
  settings = "os", "compiler", "build_type", "arch"
  generators = "cmake"

  def build(self):
    cmake = CMake(self)
    cmake.configure()
    cmake.build()

  def test(self):
    bin_path = os.path.join("bin", "test_package")
    self.run(bin_path, run_environment=True)
    self.run("flatc --version", run_environment=True)
    self.run("flathash fnv1_16 conan", run_environment=True)
