# automatically generated by the FlatBuffers compiler, do not modify

# namespace: reflection

# New schema language features that are not supported by old code generators.
class AdvancedFeatures(object):
    AdvancedArrayFeatures = 1
    AdvancedUnionFeatures = 2
    OptionalScalars = 4
    DefaultVectorsAndStrings = 8
